@@ -5,12 +5,12 @@ use std::{
     collections::HashMap,
     io,
     sync::{
-        atomic::{AtomicU32, Ordering},
+        atomic::{AtomicBool, AtomicU32, Ordering},
         Arc,
     },
 };
 
-use async_channel;
+extern crate async_channel;
 
 use futures::StreamExt;
 use futures_util::sink::SinkExt;
@@ -42,6 +42,7 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> StreamMultiplexor<T> {
         let (send, recv) = mpsc::channel(10);
         let inner = Arc::from(StreamMultiplexorInner {
             config,
+            connected: AtomicBool::from(true),
             port_connections: RwLock::from(HashMap::new()),
             port_listeners: RwLock::from(HashMap::new()),
             send: RwLock::from(send),
@@ -53,11 +54,13 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> StreamMultiplexor<T> {
             loop {
                 let frame = match recv.recv().await {
                     Some(v) => v,
-                    _ => {
-                        return;
+                    error => {
+                        error!("Error {:?} reading from stream", error);
+                        break;
                     }
                 };
-                if let Err(_) = framed_writer.send(frame).await {
+                if let Err(error) = framed_writer.send(frame).await {
+                    error!("Error {:?} reading from stream", error);
                     break;
                 }
             }
@@ -69,8 +72,21 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> StreamMultiplexor<T> {
             loop {
                 let frame: Frame = match framed_reader.next().await {
                     Some(Ok(v)) => v,
-                    _ => {
-                        return;
+                    error => {
+                        error!("Error {:?} reading from framed_reader", error);
+
+                        inner_clone.connected.store(false, Ordering::Relaxed);
+
+                        for (_, connection) in inner_clone.port_connections.write().await.drain() {
+                            if let Some(sender) = connection.external_stream_sender.write().await.as_ref() {
+                                if let Err(error) = sender.send(Err(io::Error::from(io::ErrorKind::BrokenPipe))).await {
+                                    error!("Error {:?} dropping port_connections", error);
+                                }
+                            }
+                        }
+                        inner_clone.port_listeners.write().await.clear();
+
+                        break;
                     }
                 };
                 if matches!(frame.flag, Flag::Syn)
@@ -111,15 +127,16 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> StreamMultiplexor<T> {
     }
 
     pub async fn bind(&self, port: u16) -> Result<MuxListener<T>> {
+        if !self.inner.connected.load(Ordering::Relaxed) {
+            return Err(io::Error::from(io::ErrorKind::ConnectionReset));
+        }
         let mut port = port;
         if port == 0 {
             while port < 1024 || self.inner.port_listeners.read().await.contains_key(&port) {
                 port = rand::thread_rng().gen();
             }
-        } else {
-            if self.inner.port_listeners.read().await.contains_key(&port) {
-                return Err(io::Error::from(io::ErrorKind::AddrInUse));
-            }
+        } else if self.inner.port_listeners.read().await.contains_key(&port) {
+            return Err(io::Error::from(io::ErrorKind::AddrInUse));
         }
         let (send, recv) = async_channel::bounded(8);
         self.inner.port_listeners.write().await.insert(port, send);
@@ -131,6 +148,9 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> StreamMultiplexor<T> {
     }
 
     pub async fn connect(&self, port: u16) -> Result<DuplexStream> {
+        if !self.inner.connected.load(Ordering::Relaxed) {
+            return Err(io::Error::from(io::ErrorKind::ConnectionReset));
+        }
         let mut local_port = 0;
         while local_port < 1024
             || self
@@ -144,7 +164,7 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> StreamMultiplexor<T> {
         }
 
         let mux_socket = MuxSocket::new(self.inner.clone(), local_port, port, false);
-        let rx = mux_socket.stream().await;
+        let mut rx = mux_socket.stream().await;
         self.inner
             .port_connections
             .write()
@@ -154,7 +174,7 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> StreamMultiplexor<T> {
 
         rx.recv()
             .await
-            .map_err(|e| io::Error::new(io::ErrorKind::Other, e))?
+            .ok_or(io::Error::from(io::ErrorKind::Other))?
     }
 }
 
@@ -166,10 +186,13 @@ enum PortState {
     Open,
 }
 
+type PortPair = (u16, u16);
+
 #[derive(Debug)]
 struct StreamMultiplexorInner<T> {
     config: Config,
-    port_connections: RwLock<HashMap<(u16, u16), Arc<MuxSocket<T>>>>,
+    connected: AtomicBool,
+    port_connections: RwLock<HashMap<PortPair, Arc<MuxSocket<T>>>>,
     port_listeners: RwLock<HashMap<u16, async_channel::Sender<DuplexStream>>>,
     send: RwLock<mpsc::Sender<Frame>>,
 }
@@ -182,8 +205,8 @@ pub struct MuxListener<T> {
 }
 
 impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> MuxListener<T> {
-    pub async fn accept(&self) -> Option<DuplexStream> {
-        self.recv.recv().await.ok()
+    pub async fn accept(&self) -> Result<DuplexStream> {
+        Ok(self.recv.recv().await.map_err(|e| io::Error::new(io::ErrorKind::Other, e))?)
     }
 }
 
@@ -196,7 +219,7 @@ struct MuxSocket<T> {
     state: RwLock<PortState>,
     seq: AtomicU32,
     write_half: RwLock<Option<WriteHalf<DuplexStream>>>,
-    external_stream_sender: RwLock<Option<async_channel::Sender<Result<DuplexStream>>>>,
+    external_stream_sender: RwLock<Option<mpsc::Sender<Result<DuplexStream>>>>,
 }
 
 impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> MuxSocket<T> {
@@ -218,8 +241,8 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> MuxSocket<T> {
         })
     }
 
-    pub async fn stream(self: &Arc<Self>) -> async_channel::Receiver<Result<DuplexStream>> {
-        let (sender, receiver) = async_channel::bounded(1);
+    pub async fn stream(self: &Arc<Self>) -> mpsc::Receiver<Result<DuplexStream>> {
+        let (sender, receiver) = mpsc::channel(1);
         *self.external_stream_sender.write().await = Some(sender);
         receiver
     }
@@ -310,8 +333,8 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> MuxSocket<T> {
     pub async fn recv_frame(self: &Arc<Self>, frame: Frame) {
         let state: PortState = *self.state.read().await;
         match frame.flag {
-            Flag::Syn => match state {
-                PortState::Closed => {
+            Flag::Syn => {
+                if let PortState::Closed = state {
                     if let Err(error) = self
                         .inner
                         .send
@@ -328,8 +351,7 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> MuxSocket<T> {
                     }
                     *self.state.write().await = PortState::SynAck;
                 }
-                _ => {}
-            },
+            }
             Flag::SynAck | Flag::Ack => match state {
                 PortState::SynAck | PortState::Ack => {
                     if let Err(error) = self
@@ -356,29 +378,27 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> MuxSocket<T> {
                                 error!("Error {:?} sending DuplexStream to acceptor", error);
                             }
                         }
-                    } else {
-                        if let Some(sender) = self.external_stream_sender.write().await.as_ref() {
-                            let stream = self.spawn_stream().await;
-                            if let Err(error) = sender.send(Ok(stream)).await {
-                                error!("Error {:?} sending DuplexStream to connector", error);
-                            }
+                    } else if let Some(sender) = self.external_stream_sender.write().await.as_ref()
+                    {
+                        let stream = self.spawn_stream().await;
+                        if let Err(error) = sender.send(Ok(stream)).await {
+                            error!("Error {:?} sending DuplexStream to connector", error);
                         }
                     }
                 }
                 _ => {}
             },
-            Flag::Unset => match state {
-                PortState::Open => {
+            Flag::Unset => {
+                if let PortState::Open = state {
                     if let Some(write_half) = self.write_half.write().await.as_mut() {
                         if let Err(error) = write_half.write_all(&frame.data).await {
                             error!("Error {:?} writing data to write_half", error);
                         }
                     }
                 }
-                _ => {}
-            },
-            Flag::Fin => match state {
-                PortState::Open => {
+            }
+            Flag::Fin => {
+                if let PortState::Open = state {
                     if let Err(error) = self
                         .inner
                         .send
@@ -396,15 +416,10 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> MuxSocket<T> {
                     *self.state.write().await = PortState::Closed;
                     *self.write_half.write().await = None;
                 }
-                _ => {}
-            },
+            }
             Flag::Rst => {
                 if matches!(*self.state.read().await, PortState::Closed | PortState::Ack) {
-                    if let Some(stream_sender) = self
-                        .external_stream_sender
-                        .write()
-                        .await
-                        .as_ref()
+                    if let Some(stream_sender) = self.external_stream_sender.write().await.as_ref()
                     {
                         if let Err(error) = stream_sender
                             .send(Err(io::Error::from(io::ErrorKind::AddrNotAvailable)))
@@ -421,6 +436,12 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> MuxSocket<T> {
     }
 }
 
+#[cfg(test)]
+#[ctor::ctor]
+fn init_tests() {
+    tracing_subscriber::fmt::init();
+}
+
 #[tokio::test]
 async fn connect_no_listen_fails() {
     let (a, b) = duplex(10);
@@ -454,7 +475,7 @@ async fn dropped_connection_rsts() {
     let sm_b = StreamMultiplexor::new(b, Config::default());
 
     tokio::spawn(async move {
-        sm_b.bind(22).await.unwrap().accept().await;
+        sm_b.bind(22).await.unwrap().accept().await.unwrap();
     });
 
     assert!(matches!(sm_a.connect(22).await, Ok(_)));
@@ -492,3 +513,44 @@ async fn connected_stream_passes_data() {
 
     assert_eq!(input_bytes, output_bytes);
 }
+
+#[tokio::test]
+async fn wrapped_stream_disconnect() {
+    use tokio::net::{TcpListener, TcpStream};
+
+    let listener = TcpListener::bind("127.0.0.1:0").await.unwrap();
+    let local_addr = listener.local_addr().unwrap();
+
+    tokio::spawn(async move {
+        listener.accept().await.unwrap();
+    });
+
+    let stream = TcpStream::connect(local_addr).await.unwrap();
+
+    let sm_a = StreamMultiplexor::new(stream, Config::default());
+    tokio::time::sleep(tokio::time::Duration::from_millis(100)).await;
+
+    assert!(matches!(sm_a.connect(1024).await, Err(..)));
+    assert!(matches!(sm_a.bind(1024).await, Err(..)));
+}
+
+#[tokio::test]
+async fn wrapped_stream_disconnect_listener() {
+    use tokio::net::{TcpListener, TcpStream};
+
+    let listener = TcpListener::bind("127.0.0.1:0").await.unwrap();
+    let local_addr = listener.local_addr().unwrap();
+
+    tokio::spawn(async move {
+        let _ = listener.accept().await.unwrap();
+        tokio::time::sleep(tokio::time::Duration::from_millis(100)).await;
+    });
+
+    let stream = TcpStream::connect(local_addr).await.unwrap();
+    let sm_a = StreamMultiplexor::new(stream, Config::default());
+    let listener = sm_a.bind(1024).await.unwrap();
+
+    tokio::time::sleep(tokio::time::Duration::from_millis(100)).await;
+
+    assert!(matches!(listener.accept().await, Err(..)));
+}
